import argparse
import os

from swip.commands.add import add
from swip.commands.branch import branch
from swip.commands.checkout import checkout
from swip.commands.commit import commit
from swip.commands.init import init
from swip.commands.log import log
from swip.commands.merge import merge 
from swip.commands.status import status
from swip.common.exceptions import NoSwipError
from swip.common.helpers import logger
from swip.common.paths import create_swip_paths, find_swip_repo



parser = argparse.ArgumentParser()
subparser = parser.add_subparsers(dest='command', required=True)

# Init
init_parser = subparser.add_parser('init')

# Add
add_parser = subparser.add_parser('add')
add_parser.add_argument('path', type=str, help='a path to a file or a directory')

# Commit
commit_parser = subparser.add_parser('commit')
commit_parser.add_argument('message', type=str)

# Status
status_parser = subparser.add_parser('status')

# Branch
branch_parser = subparser.add_parser('branch')
branch_parser.add_argument('-name', '-n', metavar='<branch name>', type=str) 

# Checkout
checkout_parser = subparser.add_parser('checkout')
checkout_parser.add_argument('destination', metavar='[<branch name> | <commit id>]', type=str)

# Merge
merge_parser = subparser.add_parser('merge')
merge_parser.add_argument('branch_name', metavar='[<branch name>', type=str)

# Log
log_parser = subparser.add_parser('log')
log_parser.add_argument('-graph', action='store_true')
log_parser.add_argument('-all', action='store_true')


args = parser.parse_args()


COMMANDS = {
    'init': init,
    'add': add,
    'commit': commit,
    'status': status,
    'branch': branch,
    'checkout': checkout,
    'merge': merge,
    'log': log
}


def main():
    arg = args.command
    func = COMMANDS[arg]
    params = vars(args)
    command = params.pop("command")       
    
    if not command == 'init':
        cwd = os.getcwd()
        try:
            swip = find_swip_repo(cwd)
        except NoSwipError as err:
            logger.error(err)
            return

        create_swip_paths(swip)
    func(**params)


if __name__ == '__main__':
    main()
   


from typing import Optional

import swip.common.helpers as helpers
from swip.common.exceptions import BranchExistsError, CommitRequiredError
from swip.common.paths import PATHS as path_to
from swip.common.helpers import logger
from termcolor import colored


def validate_new_branch_name(name: str) -> None:  
    """Given a branch name, reads and iterates over
    the references file. If a matching existing branch 
    name is found, BranchExistsError will be raised
    and the branch command will fail.
    """
    with open(path_to.references, 'r') as file:
        ref_text = file.readlines()
    for line in ref_text:
        branch_name, _ = line.split('=')
        if name == branch_name:
            raise BranchExistsError(f'Branch {name} already exists')


def print_existing_branches() -> None:
    """Prints all existing branch names."""
    branches = helpers.get_existing_branches()
    active = helpers.get_active_branch()
    avail_branches = ''
    for branch in branches:
        if branch == active:
            branch = colored(f'*{branch}', 'cyan')
        avail_branches += f'{branch}\n'
    print(avail_branches)


def add_branch_to_ref(name: str, commit_id: str) -> None:
    """Adds branch name to references file with
    a pointer to the current (HEAD) commit id.
    """
    with open(path_to.references, "a") as file:
        file.write(f"\n{name}={commit_id}")


def branch(name: Optional[str] = None) -> bool:
    """If name was given, a branch with that name is created.
    Otherwise, all branches will be printed.
    """ 
    try:
        helpers.validate_ref_file_exists()
    except CommitRequiredError as err:
        logger.warning(err)
        return False

    if not name:
        print_existing_branches()
        return True

    try:
        validate_new_branch_name(name)
    except BranchExistsError as err:
        logger.warning(err)
        return False

    cur_commit_id = helpers.get_head_commit_id() 
    add_branch_to_ref(name, cur_commit_id) 
    logger.info(f'Branch {name} created')
    return True



import filecmp
import shutil
import sys
from pathlib import Path
from typing import Dict, List, Optional, Set

from loguru import logger

from swip.common.exceptions import CommitRequiredError
from swip.common.paths import PATHS as path_to


HEAD = 'HEAD'


def is_working_tree_clean(cur_status):
    """Returns False if there are changes to be commited 
    or changes not staged for commit. Returns True otherwise.
    """
    return not cur_status.to_commit and not cur_status.not_staged


# Commits 
def get_head_commit_id() -> str:
    """Return the commit id HEAD points to."""
    with open(path_to.references, 'r') as file:
        head_line = file.readline() 
    _, commit_id = head_line.split('=') 
    return commit_id.strip()   


def get_commit_by_branch(input_branch: str) -> Optional[str]: 
    """Given a branch name, returns it's latest
    commit id from the references file if branch
    exists. Otherwise, returns None.
    """
    with open(path_to.references, "r") as file:
        ref_txt = file.readlines()
    for line in ref_txt: 
        branch_name, commit_id = line.strip().split('=')
        if input_branch == branch_name:
            return commit_id
    return 


def get_all_branches_commits() -> Dict[str, str]:
    """Reads references and returns a dictionary containing
    key: value pairs of branch name: last commit id.
    """
    branches_commits = {}
    with open(path_to.references, 'r') as ref_file:
        ref_text = ref_file.readlines()
    for line in ref_text[1:]:
        branch, commit = line.strip().split('=')
        branches_commits[branch] = commit
    return branches_commits


# Refernces file
def create_references_file(commit_id: str) -> None:
    """Creates the references text file in the swip repository, which 
    will contain all existing branches and the current commit
    they point to, as well as the HEAD current commit. Each line 
    represents a specific branch in the format <branch name>=<commit id>.
    """
    with open(path_to.references, "w") as file:
        to_write = [f"HEAD={commit_id}\n", f"master={commit_id}"]
        file.writelines(to_write)


def validate_ref_file_exists():
    """Checks if references file exists. If not, 
    CommitRequiredError is raised."""
    if not path_to.references.exists():
        raise CommitRequiredError(
            "Can't be executed before first commit"
        )


def update_ref_head(ref, commit_id: str) -> None: 
    with open(ref, "r") as file:
        ref_txt = file.readlines()
        _, current_commit_id = ref_txt[0].strip().split('=')
        ref_txt[0] = ref_txt[0].replace(current_commit_id, commit_id)
    
    with open(ref, "w") as file:
        file.writelines(ref_txt)
        

# Branches
def get_active_branch() -> str:
    """Reads activated file and returns current
    activated branch.
    """
    active_branch = path_to.activated.read_text()
    return active_branch


def get_existing_branches() -> List[str]:
    """Return a list of all existing branches."""
    branches_info = get_all_branches_commits()
    branches_names = branches_info.keys()
    return branches_names


# Files
def get_rel_paths(
    dir: Path, only_files: bool = False, exclude_swip: bool = False
    ) -> Set[Path]:
    """Given a path to a directory, returns all its content paths, 
    relative to the given directory.

    If only_files is True, only files paths are returned. exclude_swip
    is used on the working directory. If True, returns all working 
    directory files excluding the swip repository content.
    """
    paths = set(dir.rglob('*'))
    if exclude_swip and only_files:
        return {p.relative_to(dir) for p in paths if 
        not p.is_relative_to(path_to.swip) and p.is_file()}
    
    if exclude_swip:
        return {p.relative_to(dir) for p in paths if 
        not p.is_relative_to(path_to.swip)}

    if only_files:
        return {p.relative_to(dir) for p in paths if p.is_file()}

    return {p.relative_to(dir) for p in paths}


def get_diff_content_files(
    mutual_files: Set[Path], dir_one_path: Path, dir_two_path: Path
    ) -> Set[Path]:
    """Given a set of mutual files which appear in both directories, 
    joins the relative path of each file to each directory path and
    compares its content. Return a set of files with different content.
    """
    changed_files = set()
    for file_path in mutual_files:
        file_one = dir_one_path / file_path
        file_two = dir_two_path / file_path
        if not filecmp.cmp(file_one, file_two, shallow=False):
            changed_files.add(file_path)
    return changed_files
    

def copy_file_to_staging(source_dir: Path, file: Path) -> None: 
    """Given a file's path, file will be copied
    to the staging area of the swip repository.
    """
    destination = path_to.staging_area / file
    destination_dir = destination.parent
    if not destination_dir.exists():
        destination_dir.mkdir(parents=True)
    abs_path = source_dir / file
    shutil.copy(abs_path, destination)


def update_working_dir(
    source_dir: Path, paths_to_keep: Optional[Set[str]] = None
    ) -> None: 
    """Given a source directory - such as staging area or
    a commit image -  updates the working dir to match the
    content of the source dir without altering or deleting
    the swip repsitory.
    
    If paths_to_keep is given, these files and dirs paths will
    also be ignored in the updating process.
    """
    working_dir_paths = get_rel_paths(path_to.working_dir, exclude_swip=True)
    if paths_to_keep:
        working_dir_paths = working_dir_paths - paths_to_keep
    for path in working_dir_paths:
        path = path_to.working_dir / path 
        if path.is_dir():
            shutil.rmtree(path, ignore_errors=False)
        elif path.is_file():
            path.unlink()
    shutil.copytree(source_dir, path_to.working_dir, dirs_exist_ok=True)


# logger formatting 
logger.remove()
logger.add(sys.stdout, format='{time:%H:%M:%S} | {module}: <lvl><n>{message}</></>')



import shutil
from pathlib import Path
from typing import Set, Tuple

import swip.common.helpers as helpers
from swip.common.exceptions import CommitRequiredError, MergeError, NoBranchError
from swip.common.paths import PATHS as path_to
from swip.common.helpers import logger

import swip.commands.log
from swip.commands.commit import commit
from swip.commands.status import get_status_info, print_status


def validate_merge(branch_name):
    """Merge command will fail if one ot these
    conditions are met:
    - Given branch name doesn't match an existing branch,.
    - There are changes to be commited or changes not staged.
    """
    if branch_name not in helpers.get_all_branches_commits():
        raise NoBranchError('No such existing branch')

    status = get_status_info()
    if not helpers.is_working_tree_clean(status):
        print_status(status)
        raise MergeError('In order to Merge, please clear your working tree')
 
         
def find_mutual_ancestor(head_commit, branch_commit):
    """Given two commit ids, builds each commit's parents
    history and finds the first mutual commit of the two
    histories. 
    """
    active_history = []
    branch_history = []
    commands.log.get_branch_history(head_commit, active_history)
    commands.log.get_branch_history(branch_commit, branch_history)
    active_parents =  [commit['commit'] for commit in active_history]
    branch_parents = [commit['commit'] for commit in branch_history]
    for commit in branch_parents:
        if commit in active_parents:
            return commit
    return False
            

def compare_dir_files(
    branch_dir: Path, ancestor_dir: Path
    ) -> Tuple[Set[Path], Set[Path]]:
    """Returns files that were added to a branch and
    do not appear in the active branch, and the files
    that exist in both branches.
    """
    branch_files = helpers.get_rel_paths(branch_dir, only_files=True) 
    ancestor_files = helpers.get_rel_paths(ancestor_dir, only_files=True) 
    added_files = branch_files - ancestor_files
    mutual_files = branch_files & ancestor_files
    changed_files = helpers.get_diff_content_files(mutual_files, branch_dir, ancestor_dir) 
    return changed_files, added_files


def copy_changed_files(changed_files: Set[Path], branch_dir: Path) -> None:
    """Copies all changed files from the merged branch
    last commit to the staging area.
    """
    for file in changed_files:
        source_path = branch_dir / file
        dest_path = path_to.staging_area / file
        shutil.copy(source_path, dest_path)


def update_staging_area(
    branch_dir: Path, added_files: Set[Path], changed_files: Set[Path]
    ) -> None:
    """Copies all changed files and added files 
    (including missing parents) to staging area.
    """
    files_to_copy = added_files + changed_files
    for file in files_to_copy:
        helpers.copy_file_to_staging(branch_dir, file) 
    copy_changed_files(changed_files, branch_dir)


def merge_commit(branch_name: str, head_commit: str, branch_commit: str) -> None:
    """Executes a merge-commit: a commit
    that has two parents. The commit message
    indicates that it's a merge-commit.
    """
    active_branch = helpers.get_active_branch()
    merge_message = f'{branch_name} Merged into {active_branch} (HEAD)'
    parents = f'{head_commit},{branch_commit}'
    commit(merge_message, parents)


def merge(branch_name: str) -> bool:
    """Merges given branch name into current active branch."""
    try:
        validate_merge(branch_name)
    except (CommitRequiredError, MergeError, NoBranchError) as err:
        logger.warning(err)
        return False

    head_commit = helpers.get_head_commit_id() 
    branch_commit = helpers.get_commit_by_branch(branch_name)
    mutual_ancestor = find_mutual_ancestor(head_commit, branch_commit)
    branch_dir = path_to.images / branch_commit 
    ancestor_dir = path_to.images / mutual_ancestor
    changed_files, added_files = compare_dir_files(branch_dir, ancestor_dir)
    update_staging_area(branch_dir, added_files, changed_files)
    merge_commit(branch_name, head_commit, branch_commit)
    helpers.update_working_dir(path_to.staging_area) # what if there is untracked files?
    return True





import shutil
from collections import namedtuple
from pathlib import Path
from typing import Optional, Set, Tuple

import swip.common.helpers as helpers
from swip.common.exceptions import CheckoutError, CommitRequiredError
from swip.common.paths import PATHS as path_to
from swip.common.helpers import logger

from swip.commands.status import get_status_info, print_status


def is_valid_branch(destination: str) -> Optional[str]:
    """Given the user input, checks if it matches a 
    branch name. If so, returns the commit id of the
    branch. Otherwise, returns None.
    """
    commit_id = helpers.get_commit_by_branch(destination)
    return None if not commit_id else commit_id

    
def verify_commit_id(commit_id: str) -> bool:
    """Returns True if commit_id matches an existing
    commit. Otherwise, returns False.
    """
    commit_dir_path = path_to.images / commit_id
    return commit_dir_path.exists()


def validate_checkout(status: namedtuple, destination: str) -> Tuple[str, str]:
    """Checkout command will fail if one of these
    conditions are met:
    - There are changes to be commited or changes not staged.
    - User destination input doesn't match a branch name or 
    a commit id known to Swip.

    If a branch name was given, returns a tuple of branch name
    and its HEAD commit id. If branch wasn't given, branch 
    name will be None. On failure, CheckoutError will be thrown.
    """
    branch = None
    if not helpers.is_working_tree_clean(status):
        print_status(status)
        raise CheckoutError(
            'In order to Checkout, pleae clear your working tree'
        )
    
    commit_by_branch = is_valid_branch(destination)
    if commit_by_branch:
        branch = destination
        commit_id = commit_by_branch
        return branch, commit_id

    if not verify_commit_id(destination):
        raise CheckoutError("Input doesn't match any commit or branch")
    
    commit_id = destination
    return branch, commit_id


def update_staging_area(commit_dir: Path) -> None:
    """Removes all staging area content and replaces it
    with the given commit directory content.
    """
    shutil.rmtree(path_to.staging_area, ignore_errors=False)
    shutil.copytree(commit_dir, path_to.staging_area, dirs_exist_ok=True)


def paths_to_ignore(untracked_files: Set[Optional[Path]]) -> Set[Optional[Path]]: 
    """Given a Set of untracked files, returns the
    files' paths, including parent directories, to ignore. 
    """
    working_dir = path_to.working_dir
    paths_to_ignore = set()
    for path in untracked_files:
        path = working_dir / path
        while path != working_dir:
            paths_to_ignore.add(path) 
            path = path.parent
    return paths_to_ignore  



def update_active_branch(branch_name: str) -> None:
    """Overwrites activated file to contain the given branch,
    which will become the active branch.
    """
    path_to.activated.write_text(branch_name)
            

def checkout(destination: str) -> bool:
    """Checks out a branch or a commit id.
    The given content of the commit id or HEAD of
    given branch will be copied to the working
    directory and staging area.
    """
    try:
        status = get_status_info()
    except CommitRequiredError as err:
        logger.warning(err)
        return False

    try:
        branch, commit_id = validate_checkout(status, destination)
    except CheckoutError as err:
        logger.warning(err)
        return False

    commit_dir = path_to.images/ commit_id
    update_staging_area(commit_dir)
    untracked_paths = paths_to_ignore(status.untracked_files) 
    helpers.update_working_dir(commit_dir, paths_to_keep=untracked_paths)

    if branch:
        update_active_branch(branch)
    helpers.update_ref_head(commit_id)
    logger.info(f'Switched to {destination}') 
    return True   
    



class SwipError(Exception):
    pass


class NoSwipError(SwipError):
    pass


class CommitError(SwipError):
    pass


class CommitRequiredError(FileNotFoundError):
    pass


class BranchExistsError(SwipError):
    pass  


class NoBranchError(SwipError):
    pass


class CheckoutError(SwipError):
    pass


class MergeError(SwipError):
    pass



from pathlib import Path

from swip.common.exceptions import NoSwipError


def find_swip_repo(given_path):  
    path = Path(given_path)
    swip_path = None
    if path.is_file():
        path = path.parent

    while path != path.parent and swip_path is None:
        subdirs = [p for p in path.iterdir() if p.is_dir()] 
        for dir in subdirs:
            if dir.name == ".swip":
                swip_path = dir
                return swip_path
        path = path.parent
    raise NoSwipError(
        'No .Swip repository in this directory or any of its parents'
        )


class SwipPaths:
    swip: Path
    working_dir: Path
    images: Path
    staging_area: Path
    references: Path
    activated: Path


PATHS = SwipPaths()

def create_swip_paths(swip):
    PATHS.swip = swip
    PATHS.working_dir = swip.parent
    PATHS.images = swip / 'images'
    PATHS.staging_area = swip / 'staging_area'
    PATHS.references = swip / 'references.txt'
    PATHS.activated = swip / 'activated.txt'



from typing import Dict, List, Optional, Tuple

import graphviz
import networkx as nx
from matplotlib import pyplot as plt

from swip.common.paths import PATHS as path_to
import swip.common.helpers as helpers
from swip.common.exceptions import CommitRequiredError
from swip.common.helpers import logger, HEAD


def get_branch_history(head_commit: str, history: List) -> Optional[Dict[str, str]]:
    """Returns a list of dictionaries containing the history 
    of a given branch. Each record contains a commit id, the 
    date and time, the commit message and the commit's parent(s).
    """
    if head_commit == 'None':
        return

    for res in history:
        if res['commit'] == head_commit:
            return

    commit_file = path_to.images.joinpath(head_commit + '.txt')    
    with open(commit_file, 'r') as commit_txt:
        lines = commit_txt.readlines()

    res = {}
    res['commit'] = head_commit
    _, parents = lines[0].strip().split("=")  
    res['date'] = lines[1].strip().split("=")[1]
    res['message'] = lines[2].strip().split("=")[1]
    
    if ',' not in parents:
        res['parents'] = [parents]
        history.append(res)
        return get_branch_history(parents, history)

    parent1, parent2 = parents.split(',')
    res['parents'] = [parent1, parent2]
    history.append(res)
    get_branch_history(parent2.strip(), history)
    return get_branch_history(parent1.strip(), history)


def get_log_history(
    all: bool, commits: List[str], head_commit: str
    ) -> List[Dict[str, str]]:
    """Given all is True, returns a history information
    of all existing branches . Otherwise, returns only
    the history of the current active branch. 
    """
    history = []
    if all:
        for commit in commits:
            get_branch_history(commit, history)
    else:
        get_branch_history(head_commit, history)
    return history


def add_branches_pointers_to_log(
    log: List[Dict[str, str]], branches_info: Dict[str, str]
    ) -> None:
    """Given branch history list returned by get_branch_history,
    and a dictionary containing branches and their respective
    commit ids, adds to the commit's value in the branch history
    the pointed branch's name and HEAD.
    """
    active_branch = helpers.get_active_branch()
    for branch, commit in branches_info.items():
        if active_branch == branch:
            branch = f'{HEAD} -> {branch}'
        for dict in log:
            if dict['commit'] == commit:
                dict['commit'] += f' ({branch})'
 

def print_log(log: List[Dict[str, str]]) -> str:
    """Iterate over the log list of dictionaries, where each 
    dictionary represents a commit id info, to build a string
    containing all the commits information.
    """
    info = ""
    for image in log:
        info += f"Commit: {image['commit']}\n"
        if len(image['parents']) > 1:
            parent1, parent2 = image['parents']
            info += f"Merged: {parent1[:6]} {parent2[:6]}\n"
        info += f"Date: {image['date']}\n"
        info += f"Message: {image['message']}\n\n"
    print(info)


def get_branch_pointers(head_commit, branches_commits):
    """Returns a list of tuples containing branches and 
    their respective HEAD pointers.
    """
    pointers = []
    pointers.append((HEAD, head_commit[:6]))
    for branch, commit in branches_commits.items():
        pointers.append((branch, commit[:6]))
    return pointers


def create_graph(log: List[Dict[str, str]], branch_pointers: Tuple[str, str]) -> None:
    """Iterates over the log dictionary and creates the graph
    which will be drawn to the user based on these links.
    """
    parents = {l['commit']: l['parents'] for l in log
            if l['parents'] != ['None']}
    parents_edges = [
        (node[:6], parent[:6])
        for node, parents in parents.items()
        for parent in parents]

    # dot = graphviz.Digraph(format='svg')
    # dot.edges(parents_edges)
    # dot.edges(branch_pointers)
    # dot.render(directory='.swip/doctest-output', view=True) 
    g = nx.DiGraph()
    g.add_edges_from(parents_edges)
    g.add_edges_from(branch_pointers)
    nx.draw_networkx(g, arrows=True)
    plt.show()


   

def log(graph=False, all=False):
    """Prints a branch history. If 'all' options is used,
    the history of all branches combined is printed.
    If 'graph' is used, a graph of the history is shown
    to the user.
    """
    try:
        helpers.validate_ref_file_exists()
    except CommitRequiredError as err:
        logger.warning(err)
        return False

    head_commit = helpers.get_head_commit_id()
    branches_and_commits = helpers.get_all_branches_commits()
    branches_last_commits = branches_and_commits.values()
    history = get_log_history(all, branches_last_commits, head_commit)
    
    add_branches_pointers_to_log(history, branches_and_commits)
    print_log(history)

    if graph:
        graph_pointers = get_branch_pointers(head_commit, branches_and_commits)
        create_graph(history, graph_pointers) 
    return True
    


import random
import shutil
import string
from datetime import datetime

import swip.common.helpers as helpers
from swip.common.exceptions import CommitError
from swip.common.paths import PATHS as path_to
from swip.common.helpers import logger, HEAD

from swip.commands.status import get_status_info


def is_commit_allowed() -> None:  # Convention: returns True or False 
    """Checks the staging area directory. Throws CommitError
    if the staging directory is empty, or if no changes were
    made since last commit, causing the commit command to
    fail.
    """ 
    staging_content = list(path_to.staging_area.iterdir())
    if not staging_content:
        raise CommitError("Use the add command in order to add files to commit")    
    
    if path_to.references.exists():
        status = get_status_info()
        if not status.to_commit:  
            raise CommitError('No Files added to commit')
            

def generate_commit_id(commit_len: int = 40) -> str:
    """Returns a string consisting of 40 chars randomly
    selected from 0-9 & a-z.
    """
    chars = string.ascii_lowercase + string.digits
    commit_id = "".join(random.choice(chars) for _ in range(commit_len))
    return commit_id


def create_commit_dir(commit_id: str) -> str: 
    """Creates a new image commit directory within the images
    directory, which will contain the current staging area
    content.
    """
    commit_path = path_to.images / commit_id
    commit_path.mkdir()
    return commit_path
    

def create_image(commit_path: str) -> None:
    """Copies the content of the staging area to 
    a given commit directory path. Dirs that exists will not
    be replaced, files that exist will be overwritten.
    """
    destination = commit_path
    shutil.copytree(path_to.staging_area, destination, dirs_exist_ok=True)
    

def create_commit_metadata(commit_path: str, message: str, parents) -> None: 
    """Creates a commit metadata text file inside the images dir, 
    and writes the following information for each commit: 
    The commit's parent(s) if exists, the current time and date,
    and the commit message entered by the user. The file name
    is identical to the commit's directory name.
    """
    current_time = str(datetime.now().strftime("%a %b %w %H:%M:%S %Y"))
    if not parents:
        parents = helpers.get_head_commit_id() if path_to.references.exists() else None
    with open(f"{commit_path}.txt", "w") as file:
        to_write = [f"parent={parents}\n", f"date={current_time}\n", f"message={message}\n\n"] 
        file.writelines(to_write)


def update_references_file(commit_id):
    """Updates the HEAD & branch pointers in the references 
    file to the new commit id. If HEAD and the branch point 
    to the same commit, the branch's commit id will 
    be updated as well.
    """
    active = helpers.get_active_branch()
    head = helpers.get_head_commit_id()
    with open(path_to.references, "r") as file: 
        ref_lines = file.readlines()
    
    new_txt = ''
    for line in ref_lines:
        if line.startswith(HEAD): 
            line = line.replace(head, commit_id)
        branch_name, branch_commit = line.strip().split('=')
        if branch_commit == head and branch_name == active:
            line = line.replace(branch_commit, commit_id)
        new_txt += line
    return new_txt


def handle_references(commit_id: str) -> None:
    """Handles the update or creation of the references file.
    If it's the first commit, i.e. file does not exist,
    references file will be created. Otherwise, it will be
    updated.
    """
    if not path_to.references.exists(): 
        helpers.create_references_file(commit_id)
        return
    new_refs_txt = update_references_file(commit_id) 
    with open(path_to.references, "w") as file:
        file.write(new_refs_txt)
         

def commit(message: str, parents=None) -> bool:
    """Commits changes that were added to the staging area."""
    try:
        is_commit_allowed() 
    except CommitError as err:
        logger.warning(err)
        return False
    commit_id = generate_commit_id() 
    commit_path = create_commit_dir(commit_id)
    create_commit_metadata(commit_path, message, parents)
    create_image(commit_path) 
    handle_references(commit_id)
    logger.info(f'Commit {commit_id[:6]} created')
    return True 




import shutil
from pathlib import Path

import swip.common.helpers as helpers
from swip.common.paths import PATHS as path_to
from swip.common.helpers import logger


def validate_path(input_path):
    """Checks if the user input path matches an existing
    file. Returns the absolue path on success, otherwise
    FileNotFoundError is thrown.
    """
    abs_path = input_path.resolve()
    if not abs_path.exists(): 
        raise FileNotFoundError(
            f"{abs_path} does not match an existing file"
            )
    return abs_path


def copy_subdir_to_staging(working_dir: Path, dir: Path) -> None:
    """Given a path to a directory, it will be
    copied recursively to the staging area of the swip 
    repository. Parent directories are created if missing.
    """
    relative_path = dir.relative_to(working_dir)
    destination = path_to.staging_area / relative_path
    shutil.copytree(dir, destination, dirs_exist_ok=True)


# This is working, but repeats itself.
def copy_working_directory(working_dir: Path) -> None:
    """Given the input path is the working directory itself,
    the working dir contents will be copied, excluding
    the .swip repo and its content.
    """
    paths_to_copy = helpers.get_rel_paths(working_dir, exclude_swip=True)
    for path in paths_to_copy:
        if path.is_file():
            helpers.copy_file_to_staging(working_dir, path)
        if path.is_dir():
            path = working_dir / path
            copy_subdir_to_staging(working_dir, path)


def handle_copying_to_staging(path: Path) -> None:
    """Given a path, checks if the path is the working
    directory itself, a subdirectory, or a file. 
    For each option, the appropriate copying function 
    will be excecuted.
    """
    working_dir = path_to.working_dir
    if path == working_dir:
        copy_working_directory(working_dir)
        return
    if path.is_file():
        file_rel_path = path.relative_to(working_dir)
        helpers.copy_file_to_staging(working_dir, file_rel_path)
        return 
    copy_subdir_to_staging(working_dir, path)


def add(path: str) -> bool:
    """Adds a file or directory path to the staging area."""
    user_path = Path(path)
    try:
        abs_path = validate_path(user_path)
    except FileNotFoundError as er:
        logger.warning(er)
        return False
    handle_copying_to_staging(abs_path)
    return True


from pathlib import Path

from swip.common.helpers import logger


def create_swip_subdirs(repo_path: Path) -> None: 
    """Creates the images and staging_area subdirs in 
    the swip rpository.
    """
    for subdir in ['images', 'staging_area']:
        joined_path = repo_path / subdir
        joined_path.mkdir()


def create_activated_file(repo_path: str) -> None:
    """Creates a text file named activated, which will 
    contain the currnet active branch name.
    """
    activated_file = repo_path / 'activated.txt'
    with open(activated_file, 'w') as file:
        file.write('master')


def init(): 
    """Initializes a swip repository in project directory."""
    cwd = Path.cwd()
    swip_repo = cwd.joinpath('.swip')
    
    try:
        swip_repo.mkdir()
    except OSError:
        logger.warning(f'A Swip repository already initiated in {cwd}') 
        return False

    create_swip_subdirs(swip_repo) 
    create_activated_file(swip_repo)
    logger.info(f'Initialized in {swip_repo}. Ready to go!')
    return True



import filecmp
from pathlib import Path
from typing import NamedTuple, Optional, Set

import swip.common.helpers as helpers
from swip.common.exceptions import CommitRequiredError
from swip.common.paths import PATHS as path_to
from swip.common.helpers import logger


class Status(NamedTuple):
    active_branch: str
    to_commit: str
    not_staged: str
    untracked_files: str


# currenly no support for dirs.
# To do? this function is too similar to compare dirs in Merge.
def get_to_be_committed(last_commit_id: str) -> Set[Optional[Path]]:
    """Iterates over staging area dir in search for files that 
    are in the staging area but not in the last commit. For files 
    that are in both, a comparison will be executed to check if
    changes were made. Returns a set containing newly added and 
    changed files & dirs. 
    """ 
    commit_dir = path_to.images / last_commit_id 
    staging_paths = helpers.get_rel_paths(path_to.staging_area, only_files=True) 
    diff = set()
    for path in staging_paths:
        file_in_staging = path_to.staging_area / path
        file_in_commit = commit_dir / path
        if not file_in_commit.exists():
            diff.add(path)
        elif not filecmp.cmp(file_in_staging, file_in_commit, shallow=False):
            diff.add(path)
    return diff
        

def get_not_staged() -> Set[Optional[Path]]:
    """Finds and compares the mutual files of staging area
    and working directory. Returns a set containing the files
    in which changes were made since last commit.
    """
    diff = set()
    staging_files = helpers.get_rel_paths(path_to.staging_area, only_files=True)
    working_dir_files = helpers.get_rel_paths(path_to.working_dir, only_files=True) 
    mutual_files = staging_files & working_dir_files
    diff = helpers.get_diff_content_files(mutual_files, path_to.staging_area, path_to.working_dir)
    return diff


def get_untracked_files() -> Set[Optional[Path]]:
    """Compares the content of the working directory - excluding the swit
    repository content - and the content of the staging area. Returns a 
    set containing the files that are in the working directory but
    not in the staging area.
    """
    diff = set()
    working_dir_paths = helpers.get_rel_paths(path_to.working_dir, exclude_swip=True)
    staging_paths = helpers.get_rel_paths(path_to.staging_area)
    diff = working_dir_paths - staging_paths
    return diff


def get_status_info() -> NamedTuple: 
    """Returns a NamedTuple containing the current commit id and 
    all the status sections - changes to be committed, changed not
    staged for commit and untracked files.
    """
    try:
        cur_commit = helpers.get_head_commit_id()
    except FileNotFoundError:
        raise CommitRequiredError("Can't be executed before first commit")

    active_branch = helpers.get_active_branch()
    to_commit = get_to_be_committed(cur_commit)
    not_staged = get_not_staged()
    untracked_files = get_untracked_files()
    cur_status = Status(
        active_branch, to_commit, not_staged, untracked_files
    )
    return cur_status

 
def prettify_section(section: Set[Optional[str]]) -> str:
    """Given a set of paths, returns a string of these paths."""
    to_print = ''
    if not section:
        return ' -\n' 
    for path in section:
        to_print += f' {path}\n'
    return to_print
    

def print_status(status: NamedTuple) -> None:
    """Given current status info and current active branch, 
    prints all info to the user.
    """
    print('-' * 15)
    print(f'On branch: {status.active_branch}')
    print(f'Changes to be committed:\n{prettify_section(status.to_commit)}')
    print(f'Changes not staged:\n{prettify_section(status.not_staged)}')
    print(f'Untracked files:\n{prettify_section(status.untracked_files)}')
 

def status():
    try:
        cur_status = get_status_info()
    except CommitRequiredError as err:
        logger.warning(err)
        return False
        
    print_status(cur_status)
    return True

    


